// Copyright 2016-2020, Pulumi Corporation.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package model

import (
	"fmt"

	"github.com/hashicorp/hcl/v2"
	"github.com/hashicorp/hcl/v2/hclsyntax"
	"github.com/pulumi/pulumi/pkg/v2/codegen/hcl2/syntax"
)

// PromiseType represents eventual values that do not carry additional information.
type PromiseType struct {
	// ElementType is the element type of the promise.
	ElementType Type

	s string
}

// The set of promise types, indexed by element type.
var promiseTypes = map[Type]*PromiseType{}

// NewPromiseType creates a new promise type with the given element type after replacing any promise types within
// the element type with their respective element types.
func NewPromiseType(elementType Type) *PromiseType {
	elementType = ResolvePromises(elementType)
	if t, ok := promiseTypes[elementType]; ok {
		return t
	}

	t := &PromiseType{ElementType: elementType}
	promiseTypes[elementType] = t
	return t
}

// SyntaxNode returns the syntax node for the type. This is always syntax.None.
func (*PromiseType) SyntaxNode() hclsyntax.Node {
	return syntax.None
}

// Traverse attempts to traverse the promise type with the given traverser. The result type of traverse(promise(T))
// is promise(traverse(T)).
func (t *PromiseType) Traverse(traverser hcl.Traverser) (Traversable, hcl.Diagnostics) {
	element, diagnostics := t.ElementType.Traverse(traverser)
	return NewPromiseType(element.(Type)), diagnostics
}

// AssignableFrom returns true if this type is assignable from the indicated source type. A promise(T) is assignable
// from values of type promise(U) and U, where T is assignable from U.
func (t *PromiseType) AssignableFrom(src Type) bool {
	return assignableFrom(t, src, func() bool {
		if src, ok := src.(*PromiseType); ok {
			return t.ElementType.AssignableFrom(src.ElementType)
		}
		return t.ElementType.AssignableFrom(src)
	})
}

func (t *PromiseType) ConversionFrom(src Type) ConversionKind {
	return t.conversionFrom(src, false)
}

func (t *PromiseType) conversionFrom(src Type, unifying bool) ConversionKind {
	return conversionFrom(t, src, unifying, func() ConversionKind {
		if src, ok := src.(*PromiseType); ok {
			return t.ElementType.conversionFrom(src.ElementType, unifying)
		}
		return t.ElementType.conversionFrom(src, unifying)
	})
}

func (t *PromiseType) String() string {
	if t.s == "" {
		t.s = fmt.Sprintf("promise(%v)", t.ElementType)
	}
	return t.s
}

func (t *PromiseType) unify(other Type) (Type, ConversionKind) {
	return unify(t, other, func() (Type, ConversionKind) {
		switch other := other.(type) {
		case *PromiseType:
			// If the other type is a promise type, unify based on the element type.
			elementType, conversionKind := t.ElementType.unify(other.ElementType)
			return NewPromiseType(elementType), conversionKind
		case *OutputType:
			// If the other type is an output type, prefer the optional type, but unify the element type.
			elementType, conversionKind := t.unify(other.ElementType)
			return NewOutputType(elementType), conversionKind
		default:
			// Prefer the promise type.
			return t, t.conversionFrom(other, true)
		}
	})
}

func (t *PromiseType) isType() {}
